#!/usr/bin/perl

######################################################################
package Cognition::HTMLParser;
######################################################################

use Cognition::DocModel::Figure;
use Cognition::DocModel::Section;
use Cognition::DocModel::Table;
use Cognition::DocModel::XOXO;
use Cognition::Export::Entity;
use Cognition::HTMLParser::Utils;
use Cognition::MagicString;
use Cognition::RDFModel::Property;
use Cognition::RDFModel::Subject;
use Cognition::RDFModel::Value;
use Cognition::uF::RelEnclosure;
use Cognition::uF::RelLicense;
use Cognition::uF::RelTag;
use Cognition::uF::XFN;
use Cognition::uF::adr;
use Cognition::uF::figure;
use Cognition::uF::geo;
use Cognition::uF::hAlarm;
use Cognition::uF::hAtom;
use Cognition::uF::hCalendar;
use Cognition::uF::hCard;
use Cognition::uF::hEvent;
use Cognition::uF::hTodo;
use Cognition::uF::xoxo;
use Data::Dumper;
use Encode;
use HTML::Entities qw(:DEFAULT encode_entities_numeric %entity2char);
use HTML::TreeBuilder;
use HTTP::Request;
use RDF::Redland;
use URI::URL;
use XML::LibXML qw(:all);
use XML::LibXSLT;
use strict;

BEGIN
{
	our $VERSION = 0.1;
	our $XDP;
	$Data::Dumper::Indent = 1;
}

sub new
{
	my $class    = shift;
	my $source   = shift;
	my $opts     = shift;

	my $this   = {
		'UA'       => $opts->{ua},
		'REQUEST'  => $opts->{request},
		'RESPONSE' => $opts->{response},
		'options'  => {
			'erdf_strict_profiles'    => 1,
			'grddl_fetch'  => 0,
			'grddl_strict_profiles'   => 0,
			'p_AUTO'       => 1,
			'p_erdf'       => 1,
			'p_grddl'      => 1,
			'p_rdfx'       => 1,
			'p_rdf'        => 1,
			'p_comments'   => 1,
			'p_rdfa'       => 1,
			'p_structure'  => 1,
			'p_uf'         => 1,
			'rdfa_strict_doctype'     => 0,
			'uf_strict_profiles'      => 0
		}
	};
	bless $this, $class;
	
	$this->{BASE}    = $this->{RESPONSE}->base();
	$this->{Source} = $source; 

	foreach my $o (keys %$opts)
	{
		$this->{options}->{$o} = $opts->{$o}
			if (defined $opts->{$o} && defined $this->{options}->{$o});
	}
	
	foreach my $h (@{ $opts->{nofollow} })
	{
		$this->{options}->{nofollow}->{$h} = 1;
	}
	
	# XML::LibXML parser.
	$Cognition::HTMLParser::XDP = XML::LibXML->new();
	
	# Attempt to parse.
	if (!eval {
		$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
	})
	
	# If parsing failed, then 
	{
		$this->{_raw_source} = $this->{Source};
		
		# Figure out encoding.
		my $enc = 'iso-8859-1';
		if ($this->{RESPONSE}->header('content-type') =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i
		|| substr($this->{Source},0,2048) =~ /\bencoding=[\"\'\s]?([a-z0-9-]+)/i 
		|| substr($this->{Source},0,2048) =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i )
			{ $enc = lc($1); }
		
		# Decode into UTF-8.
		if ($enc ne 'utf-8')
		{
			$this->{decoded_from} = $enc;
			$this->{Source} = decode($enc, $this->{Source});
		}

		# Try parsing again.
		if (!eval {
			$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
		})
		
		# If it still fails, then
		{
			# Last ditch attempt at getting this into usable XHTML...
			$this->{Source} = $this->html2xhtml($this->{Source});
			
			# If we can't parse it yet, then die.
			if (!eval {
				$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
			})
			{
				print "COULD NOT PARSE!\n";
				die;
			}
		}
	}

	# These are handy.
	Cognition::HTMLParser::Utils::PrepareXPointers($this->{DOM});
	Cognition::HTMLParser::Utils::PrepareXmlLang($this->{DOM}->documentElement);

	# Parse metadata automatically if that option is enabled.
	$this->parse_metadata if ($this->{options}->{p_AUTO});

	# We're done.
	return $this;
	
} #/sub new


sub dumpText
{
	my $this = shift;
	return Dumper($this);
} #/sub dumpText


sub dumpTextClean
{
	my $this = shift;
	my $tmp  = {};
	$tmp->{DOM} = $this->{DOM};  $this->{DOM} = '...hidden...';
	# $tmp->{XP} = $this->{XP};  $this->{XP} = '...hidden...';
	
	$Data::Dumper::Sortkeys = \&_cleanKeys;
	my $txt = $this->dumpText;
	
	$this->{DOM} = $tmp->{DOM};
	# $this->{XP} = $tmp->{XP};

	return $txt;
} #/sub dumpTextClean


sub _cleanKeys
{
	my $hash = shift;
	my @keys = sort grep {/^(_subject|_id|[^_])/} keys %{ $hash };
	return \@keys;
}


sub quoted
{
	return '"'.(shift).'"';
} #/sub quoted


sub dumpRDF
{
	my $this    = shift;
	return Cognition::Export::Entity::multiRDF($this->{RDF});
} #/sub dumpRDF



sub parse_rfc2731_namespaces
# Discover namespaces for metadata conforming to RFC 2731. Note
# that eRDF uses the same format for metadata schemas, so we
# can re-use these for eRDF too.
{
	my $this = shift;
	
	# Pull out XML namespaces defined on the root node. For RDFa, etc we'll
	# also read namespaces from other nodes, but not at this point.
	foreach my $a ($this->{DOM}->getDocumentElement->getAttributes)
	{
		my $k = $a->getName;
		my $v = $a->getValue;
		next unless $k =~ /^xmlns\:/i;
		$k =~ s/^xmlns\://i;
	
		$this->mergeNS($k, $v, "The '$k' Namespace", '', 0);
	}
	
	# Find custom RFC 2731 namespaces. FOAF and other stuff goes here.
	foreach my $l ($this->{DOM}->getElementsByTagName('link'))
	{
		my $rel = $l->getAttribute('rel');
		
		if ($rel =~ /^schema[\.\:]([A-Z_][A-Z0-9_-]*)$/i)
		{
			my $prefix = $1;
			my $url    = $this->abs_url($l->getAttribute('href'));
			my $title  = $l->getAttribute('title');
			
			$this->mergeNS($prefix, $url, ($title||"The '$prefix' Schema"), undef, 1);
		}
	}

	# Well-known namespaces, some defined so that people can be lazy. Others
	# defined because they're not usually used in an RDF-context, but I thought
	# it might be nice to try.
	$this->mergeNS('dcterms', 'http://purl.org/dc/terms/', 'Dublin Core Terms', -1, 1);
	$this->mergeNS('dcmitype', 'http://purl.org/dc/dcmitype/', 'Dublin Core Types', -1, 1);
	$this->mergeNS('dc', 'http://purl.org/dc/elements/1.1/', 'Dublin Core 1.1', -1, 1);
	$this->mergeNS('foaf', 'http://xmlns.com/foaf/0.1/', 'Friend of a Friend', -1, 1);
	$this->mergeNS('http11', 'urn:ietf:rfc:2616#', 'Hypertext Transfer Protocol 1.1', -1, 1);
	$this->mergeNS('openid', 'http://openid.net/specs/openid-authentication-1_1.html#', 'OpenID 1.1', -1, 1);
	$this->mergeNS('owl', 'http://www.w3.org/2002/07/owl#', 'OWL Web Ontology Language', -1, 1);
	$this->mergeNS('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'RDF/XML Syntax Specification (Revised)', -1, 1);
	$this->mergeNS('rdfs', 'http://www.w3.org/2000/01/rdf-schema#', 'RDF Vocabulary Description Language 1.0', -1, 1);
	$this->mergeNS('schema', 'urn:ietf:rfc:2731#', "The 'Schema' Metaschema", -1, 1);
	$this->mergeNS('service', 'http://www.w3.org/2007/app#', 'Atom Publishing Protocol', -1, 1);
	$this->mergeNS('xhtml', 'http://www.w3.org/1999/xhtml#', 'XHTML', -1, 1);
	$this->mergeNS('xhv', 'http://www.w3.org/1999/xhtml/vocab#', 'XHTML Link Vocabulary', -1, 1);
	$this->mergeNS('xsd', 'http://www.w3.org/2001/XMLSchema#', 'XML Schema', -1, 1);
	
} #/sub parse_rfc2731_namespaces


sub parse_metadata_links
{
	my $this = shift;
	my $myurl = $this->abs_url('');
	
	# Get list of links.
	my @links   = $this->{DOM}->getElementsByTagName('link');
	my @anchors = $this->{DOM}->getElementsByTagName('a');
	push @links, @anchors;
	
	# For each link
	foreach my $l (@links)
	{
		# Subject: This page - SUPERSEDED BY FULL RFDa SUPPORT
		# my $SubjectThisPage = Cognition::RDFModel::Subject->new($myurl);

		# Subject: Linked page
		my $href = $this->abs_url($l->getAttribute('href'));
		next unless (length $href);
		my $SubjectLinkedPage;
		
		# Handle @title.
		if (length $l->getAttribute('title'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new($l->getAttribute('title'));
			$val->setDom($l);
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/title', $val);
		}

		# Handle @media.
		if (length $l->getAttribute('media'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new($l->getAttribute('media'));
			$val->setDom($l);
			$val->setVocab('http://www.w3.org/TR/CSS21/media.html#media-types');
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/format', $val);
		}

		# Handle @title.
		if (length $l->getAttribute('type'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new($l->getAttribute('type'));
			$val->setDom($l);
			$val->setVocab('http://purl.org/dc/terms/IMT');
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/format', $val);
		}

		# Handle @lang/@hreflang.
		if (length $l->getAttribute('lang')
		||  length $l->getAttribute('xml:lang')
		||  length $l->getAttribute('hreflang'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new(
				$l->getAttribute('hreflang') ||
				$l->getAttribute('xml:lang') ||
				$l->getAttribute('lang')
			);
			$val->setDom($l);
			$val->setVocab('http://purl.org/dc/terms/RFC4646');
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/language', $val);
		}

		# SUPERSEDED BY FULL RFDa SUPPORT
		## Handle @rel.
		#my $rel   = $l->getAttribute('rel');
		#if (length $rel)
		#{
		#	$rel =~ s/(^\s+|\s+$)//g;
		#	$rel =~ s/\s+/ /g;
		#	my @rel = split / /, $rel;
		#	
		#	my $eRDF = undef;
		#	$eRDF = '-' if (lc($l->tagName) eq 'a');
		#	
		#	foreach my $r (@rel)
		#	{
		#		my $nsrel = $this->_pfx2fq($r, '-', $l);
		#		
		#		my $val = Cognition::RDFModel::Value->new($href, 1);
		#		$val->setDom($l);
		#		$SubjectThisPage->addVal($nsrel, $val);
		#	}
		#}
		#
		## Handle @rev.
		#my $rev   = $l->getAttribute('rev');
		#if (length $rev)
		#{
		#	$rev =~ s/(^\s+|\s+$)//g;
		#	$rev =~ s/\s+/ /g;
		#	my @rev = split / /, $rev;
		#	
		#	my $eRDF = undef;
		#	$eRDF = '-' if (lc($l->tagName) eq 'a');
		#	
		#	$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
		#		unless ($SubjectLinkedPage);
		#
		#	foreach my $r (@rev)
		#	{
		#		my $nsrev = $this->_pfx2fq($r, '-', $l);
		#		
		#		my $val = Cognition::RDFModel::Value->new($myurl, 1);
		#		$val->setDom($l);
		#		$SubjectLinkedPage->addVal($nsrev, $val);
		#	}
		#}
		#
		#$this->mergeSubject($SubjectThisPage);
		
		$this->mergeSubject($SubjectLinkedPage);
	}
}


sub parse_metadata_metas
{
	my $this = shift;
	my $myurl = $this->abs_url('');
	
	# Parse <meta>
	foreach my $m ($this->{DOM}->getElementsByTagName('meta'))
	{
		# Value
		my $content = length($m->getAttribute('content')) ?
		                   $m->getAttribute('content') :
		                   Cognition::HTMLParser::Utils::STRINGIFY($m);

		# Subject: This page
		my $SubjectThisPage = Cognition::RDFModel::Subject->new($myurl);
		
		# @http-equiv
		my $name = $m->getAttribute('http-equiv');
		if (length $name)
		{
			my $nsname = $this->_pfx2fq('HTTP11.'.$name, undef, $m);

			my $val = Cognition::RDFModel::Value->new($content);
			$val->setDom($m);
			$SubjectThisPage->addVal($nsname, $val);
		}
		
		# @name/@property
		else
		{
			$name = $m->getAttribute('property') . ' ' . $m->getAttribute('name');
			next unless (length $name);

			$name =~ s/(^\s+|\s+$)//g;
			$name =~ s/\s+/ /g;
			my @name = split / /, $name;
			
			foreach my $n (@name)
			{
				my $nsname = $this->_pfx2fq($n, undef, $m);
				
				my $val = Cognition::RDFModel::Value->new($content);
				$val->setDom($m);
				if ($m->getAttribute('scheme'))
				{
					my $nsscheme = $this->_pfx2fq($m->getAttribute('scheme'), undef, $m);
					$val->setVocab($nsscheme);
				}
				$SubjectThisPage->addVal($nsname, $val);
			}
		}
		
		$this->mergeSubject($SubjectThisPage);
	}
}


sub parse_metadata_roles
{
	my $this = shift;
	my $myurl = $this->abs_url;
	
	# Parse @role
	foreach my $e ($this->{DOM}->getElementsByTagName('*'))
	{
		next unless (length $e->getAttribute('role'));
	
		# Subject: This page
		my $SubjectThisPage = Cognition::RDFModel::Subject->new($myurl);

		# Object: This element
		my $objuri;
		if (length $e->getAttribute('id'))
		{
			$objuri = $myurl . '#' . $e->getAttribute('id');
		}
		else
		{
			$objuri = Cognition::HTMLParser::Utils::XPointer($e, $myurl);
		}

		# Predicate: role
		my $role = $e->getAttribute('role');
		$role =~ s/(^\s+|\s+$)//g;
		$role =~ s/\s+/ /g;
		my @role = split / /, $role;
		
		foreach my $r (@role)
		{
			# Predefined roles from the XHTML vocab namespace
			if ($r =~ /^\:?( banner | contentinfo | definition | main | navigation
				| note | search | secondary | seealso )$/xi)
			{
				$r = "xhv:$1"
			}
			my $nsrole = $this->_pfx2fq($r, undef, $e);
			
			my $val = Cognition::RDFModel::Value->new($objuri, 1);
			$val->setDom($e);
			$SubjectThisPage->addVal($nsrole, $val);
		}		
		
		$this->mergeSubject($SubjectThisPage);
	}
}


sub parse_metadata_erdf
{
	my $this = shift;
	my $myurl = $this->abs_url('');

	return unless ($this->{Profiles}->{'http://purl.org/NET/erdf/profile'}
		|| !$this->{options}->{erdf_strict_profiles} );
	
	foreach my $e ($this->{DOM}->getElementsByTagName('*'))
	{
		# Value
		my $content = Cognition::HTMLParser::Utils::STRINGIFY($e);

		# Parse eRDF <URL:http://research.talis.com/2005/erdf/wiki/Main/RdfInHtml>.
		if (length ($e->getAttribute('class').$e->getAttribute('rel').$e->getAttribute('rev')))
		{
			# Find eRDF-style parent node. i.e. nearest ancestor with @id set.
			my $parent = $e->getParentNode;
			my $subject = undef;
			my $about = $myurl;
			while (!defined $subject && $parent->nodeType==XML_ELEMENT_NODE)
			{
				if ($parent->getAttribute('id'))
					{ $subject = $parent; }
				else
					{ $parent = $parent->getParentNode(); }
			}
			if ($subject)
			{
				$about .= '#' . $subject->getAttribute('id');
				if (length $subject->getAttribute('href'))
				{
					$about = $this->abs_url($subject->getAttribute('href'));
				}
			}
			my $ParentSubject = Cognition::RDFModel::Subject->new($about);

			my ($MySubject, $addMySubject);
			if (length $e->getAttribute('id'))
			{
				$MySubject = Cognition::RDFModel::Subject->new($myurl.'#'.$e->getAttribute('id'));
			}
			else
			{
				$MySubject = Cognition::RDFModel::Subject->new(Cognition::HTMLParser::Utils::XPointer($e, $this->abs_url));
			}

			# Normalise @class, @ref, @rev and arrayise.
			my @things = ();
			my $class = $e->getAttribute('class');
			$class =~ s/(^\s+|\s+$)//g;
			$class =~ s/\s+/ /g;
			my @class = split / /, $class;
			foreach my $c (@class)
				{ push @things, [('class', $c)]; }
			if ($e->tagName =~ /^(a|area|link)$/i)
			{
				$class = $e->getAttribute('rel');
				$class =~ s/(^\s+|\s+$)//g;
				$class =~ s/\s+/ /g;
				my @class = split / /, $class;
				foreach my $c (@class)
					{ push @things, [('rel', $c)]; }

				$class = $e->getAttribute('rev');
				$class =~ s/(^\s+|\s+$)//g;
				$class =~ s/\s+/ /g;
				my @class = split / /, $class;
				foreach my $c (@class)
					{ push @things, [('rev', $c)]; }
			}

			foreach my $x (@things)
			{
				my ($attr, $c) = @$x;
				
				# rdf:type stuff
				if ($attr eq 'class' && $c =~ /^\-/)
				{
					my $rdftype = $this->_pfx2fq(substr($c,1), '-', $e);
					
					if (($e->tagName eq 'a') && (length $rdftype > length $c))
					{
						my $LinkSubject = Cognition::RDFModel::Subject->new(
							$this->abs_url($e->getAttribute('href'))
						);
						my $val = Cognition::RDFModel::Value->new($rdftype, 1);
						$val->setDom($e);
						$LinkSubject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type', $val);
						$this->mergeSubject($LinkSubject);
					}
					elsif (length $rdftype > length $c)
					{
						my $val = Cognition::RDFModel::Value->new($rdftype, 1);
						$val->setDom($e);
						$MySubject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type', $val);
						$addMySubject++;
					}

					next;
				}
				
				# Only allow namespaced classes.
				my $nsrole = $this->_pfx2fq($c, '-', $e);
				next unless (length $nsrole > length $c);
				
				# Get value.
				my $valString = $e->getAttribute('title') 
					|| Cognition::HTMLParser::Utils::STRINGIFY($e);
				my $valType;
				
				# Resource if it's an <img> element.
				if ($attr eq 'class' && lc($e->tagName) eq 'img')
				{
					$valString = $this->abs_url($e->getAttribute('src'));
					$valType   = 1;
				}
				elsif ($attr eq 'rel' && $e->tagName =~ /^(area|a|link)$/i)
				{
					$valString = $this->abs_url($e->getAttribute('href'));
					$valType   = 1;
				}
				elsif ($attr eq 'rev' && $e->tagName =~ /^(area|a|link)$/i)
				{
					my $LinkSubject = Cognition::RDFModel::Subject->new(
						$this->abs_url($e->getAttribute('href'))
					);
					my $val = Cognition::RDFModel::Value->new($ParentSubject->uri, 1);
					$val->setDom($e);
					$LinkSubject->addVal($nsrole, $val);
					$this->mergeSubject($LinkSubject);
					next;
				}
				elsif ($attr eq 'class' && $e->tagName =~ /^(area|a|link)$/i)
				{
					my $LinkSubject = Cognition::RDFModel::Subject->new(
						$this->abs_url($e->getAttribute('href'))
					);
					my $val = Cognition::RDFModel::Value->new(
						Cognition::HTMLParser::Utils::STRINGIFY($e)
					);
					$val->setDom($e);
					$LinkSubject->addVal($nsrole, $val);
					$this->mergeSubject($LinkSubject);
					next;
				}
				
				# Resource if it has @id.
				if ($attr =~ /^(rel|rev)$/i && length $e->getAttribute('href'))
				{
					my $val = Cognition::RDFModel::Value->new($this->abs_url($e->getAttribute('href')), 1);
					$val->setDom($e);
					$ParentSubject->addVal($nsrole, $val);
				}
				elsif (length $e->getAttribute('id'))
				{
					my $val = Cognition::RDFModel::Value->new($MySubject->uri, 1);
					$val->setDom($e);
					$ParentSubject->addVal($nsrole, $val);
				}
				else
				{
					$ParentSubject->setDom($subject) if ($subject);
					my $val = Cognition::RDFModel::Value->new($valString, $valType);
					$val->setDom($e);
					$ParentSubject->addVal($nsrole, $val);
				}
			}
			
			$this->mergeSubject($ParentSubject) if ($ParentSubject->am_i_bovvered);
			$this->mergeSubject($MySubject)     if ($addMySubject);
			
		}
	}
}


sub parse_metadata_rdfa
# http://www.w3.org/TR/rdfa-syntax/#sec_5.5.
{
	my $this = shift;
	my $dom  = shift || $this->{DOM}->documentElement;
		
	return 0 unless ($dom->nodeType == XML_ELEMENT_NODE);
	
	# At the beginning of processing, an initial [evaluation context] is created
	my $base               = shift || $this->abs_url;
	my $parent_subject     = shift || $base;
	my $parent_object      = shift || undef;
	my $uri_mappings       = shift || {};
	my $incomplete_triples = shift || ();
	my $language           = shift || undef;
	
	# Processing begins by applying the processing rules below to the document
	# object, in the context of this initial [evaluation context]. All elements
	# in the tree are also processed according to the rules described below,
	# depth-first, although the [evaluation context] used for each set of rules
	# will be based on previous rules that may have been applied.

	my $current_element = $dom;

	# First, the local values are initialized
	my $recurse            = 1;
	my $skip_element       = 0;
	my $new_subject        = undef;
	my $current_object_resource = undef;
	my $local_uri_mappings = $uri_mappings;
	my $local_incomplete_triples = ();
	my $current_language   = $language;
	
	my $activity = 0;
	
	# Next the [current element] is parsed for [URI mapping]s and these are
	# added to the [local list of URI mappings]. Note that a [URI mapping] 
	# will simply overwrite any current mapping in the list that has the same
	# name
	foreach my $A ($current_element->getAttributes)
	{
		my $attr = $A->getName;
		my $val  = $A->getValue;
		
		if ($attr =~ /^xmlns\:(.*)$/i)
		{
			$local_uri_mappings->{$1} = $val;
			$this->mergeNS($1, $val, $1, $current_element->getAttribute('_xpath'), 0);
		}
	}
	# Mappings are provided by @xmlns. The value to be mapped is set by
	# the XML namespace prefix, and the value to map is the value of the
	# attribute - a URI. Note that the URI is not processed in any way;
	# in particular if it is a relative path it is not resolved against
	# the current [base]. Authors are advised to follow best practice
	# for using namespaces, which includes not using relative paths.
	
	# The [current element] is also parsed for any language information, and
	# if present, [current language] is set accordingly.
	# Language information can be provided using the general-purpose XML
	# attribute @xml:lang .
	if (length $current_element->getAttribute('xml:lang'))
		{ $current_language = $current_element->getAttribute('xml:lang'); }
	elsif (length $current_element->getAttribute('lang'))
		{ $current_language = $current_element->getAttribute('lang'); }
	elsif (length $current_element->getAttribute('_xmllang'))
		{ $current_language = $current_element->getAttribute('_xmllang'); }
		
	# If the [current element] contains no valid @rel or @rev URI, obtained
	# according to the section on CURIE and URI Processing, then the next step 
	# is to establish a value for [new subject]. Any of the attributes that 
	# can carry a resource can set [new subject]
	my $rel = $current_element->getAttribute('rel');
	$rel =~ s/(^\s+|\s+$)//g;
	$rel =~ s/\s+/ /g;
	my @rel = split / /, $rel;
	my @REL;
	foreach my $r (@rel)
	{
		my $R = $this->_pfx2fq($r, undef, $current_element);
		push @REL, $R if (length $R);
	}
	my $rev = $current_element->getAttribute('rev');
	$rev =~ s/(^\s+|\s+$)//g;
	$rev =~ s/\s+/ /g;
	my @rev = split / /, $rev;
	my @REV;
	foreach my $r (@rev)
	{
		my $R = $this->_pfx2fq($r, undef, $current_element);
		push @REV, $R if (length $R);
	}
	if ((!scalar @REL) && (!scalar @REV))
	{
		# [new subject] is set to the URI obtained from the first match
		# from the following rules:
		
		# by using the URI from @about, if present, obtained according to
		# the section on CURIE and URI Processing ; 
		if (length $current_element->getAttribute('about'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('about')); }
			
		# otherwise, by using the URI from @src, if present, obtained
		# according to the section on CURIE and URI Processing.
		elsif (length $current_element->getAttribute('src'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('src')); }
			
		# otherwise , by using the URI from @resource, if present, obtained
		# according to the section on CURIE and URI Processing ; 
		elsif (length $current_element->getAttribute('resource'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('resource')); }
			
		# otherwise , by using the URI from @href, if present, obtained
		# according to the section on CURIE and URI Processing.
		elsif (length $current_element->getAttribute('href'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('href')); }
			
		# If no URI is provided by a resource attribute, then the first
		# match from the following rules will apply: 
		else
		{
			# if the element is the head or body element then act as if
			# there is an empty @about present, and process it according to
			# the rule for @about, above; 
			if ($current_element->tagName eq 'head' || $current_element->tagName eq 'body')
			{
				$new_subject = $this->abs_url;
			}
			
			# if @instanceof is present, obtained according to the section
			# on CURIE and URI Processing , then [new subject] is set to be
			# a newly created [bnode]; 
			elsif (length $current_element->getAttribute('typeof')
				||  length $current_element->getAttribute('instanceof'))
			{
				$new_subject = Cognition::HTMLParser::Utils::XPointer($current_element, $this->abs_url);
				if (length $current_element->getAttribute('id'))
					{ $new_subject = $this->abs_url.'#'.$current_element->getAttribute('id'); }
			}
			
			# otherwise, if [parent object] is present, [new subject] is set
			# to that and the [skip element] flag is set to 'true'; 
			elsif ($parent_object)
			{
				$new_subject = $parent_object;
				$skip_element = 1;
			}
		}
	}
	
	# If the [current element] does contain a valid @rel or @rev URI, obtained 
	# according to the section on CURIE and URI Processing, then the next step 
	# is to establish both a value for [new subject] and a value for [current
	# object resource]:
	else
	{
		# [new subject] is set to the URI obtained from the first match
		# from the following rules: 
		
		# by using the URI from @about, if present, obtained according
		# to the section on CURIE and URI Processing; 
		if (length $current_element->getAttribute('about'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('about')); }
			
		# otherwise, by using the URI from @src, if present, obtained
		# according to the section on CURIE and URI Processing.
		elsif (length $current_element->getAttribute('src'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('src')); }

		# If no URI is provided then the first match from the following rules
		# will apply: 
		else
		{
			# if the element is the head or body element then act as if
			# there is an empty @about present, and process it according
			# to the rule for @about, above; 
			if ($current_element->tagName eq 'head' || $current_element->tagName eq 'body')
			{
				$new_subject = $this->abs_url;
			}
			
			# if @instanceof is present, obtained according to the section
			# on CURIE and URI Processing, then [new subject] is set to be a
			# newly created [bnode]; 
			elsif (length $current_element->getAttribute('typeof')
				||  length $current_element->getAttribute('instanceof'))
			{
				$new_subject = Cognition::HTMLParser::Utils::XPointer($current_element, $this->abs_url);
				if (length $current_element->getAttribute('id'))
					{ $new_subject = $this->abs_url.'#'.$current_element->getAttribute('id'); }
			}
			
			# otherwise, if [parent object] is present, [new subject] is set 
			# to that. 
			elsif ($parent_object)
			{
				$new_subject = $parent_object;
			}
		}
		
		# Then the [current object resource] is set to the URI obtained
		# from the first match from the following rules: 
		
		# by using the URI from @resource, if present, obtained according to
		# the section on CURIE and URI Processing; 
		if (length $current_element->getAttribute('resource'))
			{ $current_object_resource = $this->abs_url($current_element->getAttribute('resource')); }
			
		# otherwise, by using the URI from @href, if present, obtained according
		# to the section on CURIE and URI Processing.
		elsif (length $current_element->getAttribute('href'))
			{ $current_object_resource = $this->abs_url($current_element->getAttribute('href')); }
			
		# Note that final value of the [current object resource] will either
		# be null (from initialization), a full URI or a bnode. 
	}
	
	# If in any of the previous steps a [new subject] was set to a non-null
	# value, it is now used to provide a subject for type values
	if ($new_subject
	&& length ($current_element->getAttribute('instanceof')
		.$current_element->getAttribute('typeof')))
	{
		my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
		
		# One or more 'types' for the [ new subject ] can be set by using
		# @instanceof. If present, the attribute must contain one or more
		# URIs, obtained according to the section on URI and CURIE Processing...
	
		my $instanceof = $current_element->getAttribute('typeof');
		$instanceof = $current_element->getAttribute('instanceof')
			unless length($instanceof);
		$instanceof =~ s/\s+/ /g;
		$instanceof =~ s/(^\s|\s$)//g;
		my @instanceof = split / /, $instanceof;
		
		foreach my $instanceof (@instanceof)
		{
			# ... each of which is used to generate a triple as follows:
			#
			# subject
			#     [new subject] 
			# predicate
	    	#     http://www.w3.org/1999/02/22-rdf-syntax-ns#type 
			# object
			#     full URI of 'type' 

			my $VALUE   = Cognition::RDFModel::Value->new($this->_pfx2fq($instanceof, undef, $current_element), 1);
			$SUBJECT->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type', $VALUE);
			$activity++;
		}

		$this->mergeSubject($SUBJECT);
			
		# Note that none of this block is executed if there is no [new subject]
		# value, i.e., [new subject] remains null. 
	}
	
	# TOBY: An extension to RDFa for dealing with @title.
	if (defined $new_subject
	&&  length $current_element->getAttribute('title')
	&&  "$new_subject" ne $this->abs_url)
	{
		my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
		my $VALUE   = Cognition::RDFModel::Value->new($current_element->getAttribute('title'));
		$SUBJECT->addVal('http://www.w3.org/2000/01/rdf-schema#label', $VALUE);
		$this->mergeSubject($SUBJECT);
		$activity++;
	}
	
	# If in any of the previous steps a [current object resource] was set to
	# a non-null value, it is now used to generate triples
	if ($current_object_resource)
	{
		# Predicates for the [ current object resource ] can be set b
		# using one or both of the @rel and @rev attributes:
		#
		#    * If present, @rel will contain one or more URIs, obtained
		#      according to the section on CURIE and URI Processing each
		#      of which is used to generate a triple as follows:
		#
		#      subject
		#          [new subject] 
		#      predicate
		#          full URI 
		#      object
		#          [current object resource] 
		
		my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
		foreach my $r (@REL)
		{
			my $VALUE   = Cognition::RDFModel::Value->new($current_object_resource, 1);
			$SUBJECT->addVal($r, $VALUE);
		}
		$this->mergeSubject($SUBJECT);
		$activity++;

		#    * If present, @rev will contain one or more URIs, obtained
		#      according to the section on CURIE and URI Processing each
		#      of which is used to generate a triple as follows:
		#
		#      subject
		#          [current object resource] 
		#      predicate
		#          full URI 
		#      object
		#          [new subject] 
		
		foreach my $r (@REV)
		{
			my $SUBJECT = Cognition::RDFModel::Subject->new($current_object_resource);
			my $VALUE   = Cognition::RDFModel::Value->new($new_subject, 1);
			$SUBJECT->addVal($r, $VALUE);
			$this->mergeSubject($SUBJECT);
			$activity++;
		}
	}
	
	# If however [current object resource] was set to null, but there are 
	# predicates present, then they must be stored as [incomplete triple]s, 
	# pending the discovery of a subject that can be used as the object. Also, 
	# [current object resource] should be set to a newly created [bnode]
	elsif ((scalar @REL) || (scalar @REV))
	{
		# Predicates for [incomplete triple]s can be set by using one or
		# both of the @rel and @rev attributes:
		#
		#    * If present, @rel must contain one or more URIs, obtained 
		#      according to the section on CURIE and URI Processing each
		#      of which is added to the [local list of incomplete triples]
		#      as follows:
		#
		#      predicate
		#          full URI 
		#      direction
		#          forward 
		
		foreach my $r (@REL)
		{
			push @$local_incomplete_triples, {
				predicate => $r,
				direction => 'forward'
			};
		}
		
		#    * If present, @rev must contain one or more URIs, obtained
		#      according to the section on CURIE and URI Processing, each
		#      of which is added to the [local list of incomplete triples]
		#      as follows:
		#
		#      predicate
		#          full URI 
		#      direction
		#          reverse 
		
		foreach my $r (@REV)
		{
			push @$local_incomplete_triples, {
				predicate => $r,
				direction => 'reverse'
			};
		}
		
		$current_object_resource = Cognition::HTMLParser::Utils::BNodeURI(undef, 'Node');
	}
	
	# The next step of the iteration is to establish any [current object
	# literal
	
	my $prop = $current_element->getAttribute('property');
	$prop =~ s/(^\s+|\s+$)//g;
	$prop =~ s/\s+/ /g;
	my @prop = split / /, $prop;
	my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
	my $datatype;
	if (length $current_element->getAttribute('datatype'))
		{ $datatype = $this->_pfx2fq($current_element->getAttribute('datatype'), undef, $current_element); }
		
	foreach my $p (@prop)
	{	
		# Predicates for the [current object literal] can be set by using
		# @property. If present, one or more URIs are obtained according
		# to the section on CURIE and URI Processing and then the actual
		# literal value is obtained as follows:
		
		my $property = $this->_pfx2fq($p, undef, $current_element);
		next unless (length $property);
		my $VALUE;
		
		# as a [ plain literal ] if: 
		#
		# @content is present; 
		if (length $current_element->getAttribute('content'))
		{
			my $X = ms $current_element->getAttribute('content'), $current_element;
			$VALUE = Cognition::RDFModel::Value->new($X);
		}
		
		# or all children of the [current element] are text nodes;
		# or there are no child nodes;
		# or the body of the [ current element ] does have non-text
		#    child nodes but @datatype is present, with an empty value. 
		elsif ((!$current_element->getElementsByTagName('*')) || (defined $datatype && $datatype eq ''))
		{
			$VALUE = Cognition::RDFModel::Value->new(
				Cognition::HTMLParser::Utils::STRINGIFY($current_element)
			);
		}

		# Additionally, if there is a value for [current language] then
		# the value of the [plain literal] should include this language
		# information, as described in [RDF-CONCEPTS]. The actual literal
		# is either the value of @content (if present) or a string created
		# by concatenating the text content of each of the descendant
		# elements of the [current element] in document order. 
		
		# as an [XML literal] if:
		# 
      #    * the [current element] has any child nodes that are not simply
		#      text nodes, and @datatype is not present, or is present, but
		#      is set to rdf:XMLLiteral.
		#
	   # The value of the [XML literal] is a string created by serializing
		# to text, all nodes that are descendants of the [current element],
		# i.e., not including the element itself, and giving it a datatype of
		# rdf:XMLLiteral.
		#
		# TODO: Node itself should not be included in value. Actually, I prefer it
		# this way. :-|
		elsif ($datatype eq 'http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral'
		|| (!defined $datatype && $current_element->getElementsByTagName('*')))
		{
			# TOBY: Clean-up node.
			my $clone = $current_element->cloneNode(1);
			$clone->setAttribute('xml:lang', $clone->getAttribute('_xmllang'));
			$clone->removeAttribute('_xpath');
			$clone->removeAttribute('_xmllang');
			foreach my $kid ($clone->getElementsByTagName('*'))
			{
				$kid->removeAttribute('_xpath');
				$kid->removeAttribute('_xmllang');
			}
			$VALUE = Cognition::RDFModel::Value->new($clone->toString);
			$VALUE->setDatatype('http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral');
			$recurse = 0;
		}
		
		# as a [typed literal] if:
		#
		#     * @datatype is present, and does not have an empty value.
		#
		# The actual literal is either the value of @content (if present)
		# or a string created by concatenating the value of all descendant
		# text nodes, of the [current element] in turn. The final string
		# includes the datatype URI, as described in [RDF-CONCEPTS], which
		# will have been obtained according to the section on CURIE and URI
		# Processing.
		elsif (length $datatype)
		{
			$VALUE = Cognition::RDFModel::Value->new(
				Cognition::HTMLParser::Utils::STRINGIFY($current_element)
			);
			$VALUE->setDatatype($datatype);
		}
		
		else
		{
			die("How did we get here??\n");
		}

		# The [current object literal] is then used with each predicate to
		# generate a triple as follows:
		# 
		# subject
		#     [new subject] 
		# predicate
		#     full URI 
		# object
		#     [current object literal] 
		$SUBJECT->addVal($property, $VALUE);
		$this->mergeSubject($SUBJECT);
		$activity++;
		
		# Once the triple has been created, if the [datatype] of the
		# [current object literal] is rdf:XMLLiteral, then the [recurse]
		# flag is set to false.
		$recurse = 0
			if ($datatype eq 'http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral');
	}

	# If the [recurse] flag is 'true', all elements that are children of the 
	# [current element] are processed using the rules described here, using a 
	# new [evaluation context], initialized as follows
	my $flag = 0;
	if ($recurse)
	{
		foreach my $kid ($current_element->childNodes)
		{
			# If the [skip element] flag is 'true' then the new [evaluation context]
			# is a copy of the current context that was passed in to this level of
			# processing, with the [language] and [list of URI mappings] values
			# replaced with the local values; 
			if ($skip_element)
			{
				$flag = $this->parse_metadata_rdfa(
					$kid,
					$base,
					$parent_subject,
					$parent_object,
					$uri_mappings,
					$incomplete_triples,
					$language
				) || $flag;
			}
			
			# Otherwise, the values are: 
			else
			{
				$flag = $this->parse_metadata_rdfa(
					$kid,
					$base,
					$new_subject,
					(defined $current_object_resource ? $current_object_resource : ($new_subject ? $new_subject : $parent_subject)),
					$local_uri_mappings,
					$local_incomplete_triples,
					$current_language
				) || $flag;
			}
		}	
	}
	
	# If the [skip element] flag is 'false', and either: the previous step
	# resulted in a 'true' flag, or [new subject] was set to a non-null and
	# non-bnode value, then any [incomplete triple]s within the current context
	# should be completed:
	if (!$skip_element && ($flag || ((defined $new_subject) && ($new_subject !~ /^bnode:/))))
	{
		# Loop through list of incomplete triples...
		foreach my $it (@$incomplete_triples)
		{
			my $direction = $it->{direction};
			my $predicate = $it->{predicate};
			
			if ($direction eq 'forward')
			{
				my $SUBJECT = Cognition::RDFModel::Subject->new($parent_subject);
				my $VALUE   = Cognition::RDFModel::Value->new($new_subject, 1);
				$SUBJECT->addVal($predicate, $VALUE);
				$this->mergeSubject($SUBJECT);
				$activity++;
			}
			else
			{
				my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
				my $VALUE   = Cognition::RDFModel::Value->new($parent_subject, 1);
				$SUBJECT->addVal($predicate, $VALUE);
				$this->mergeSubject($SUBJECT);
				$activity++;
			}
		}
	}

	return 1 if ($activity || $new_subject || $flag);
	return 0;
}


sub parse_uformats
{
	my $this = shift;
	
	Cognition::uF::hCard::consume($this);
	Cognition::uF::adr::consume($this);
	Cognition::uF::geo::consume($this);
	Cognition::uF::XFN::consume($this);

	Cognition::uF::RelTag::consume($this);
	Cognition::uF::RelLicense::consume($this);
	Cognition::uF::RelEnclosure::consume($this);

	Cognition::uF::hCalendar::consume($this);
	Cognition::uF::hAtom::consume($this);
	
	# hReview ?
	# hListing ?
	# hResume ?
	
} #/sub parse_uformats


sub parse_doc_structure
{

	my $this = shift;
	my $rv   = { kids=>() };
	my $lvl  = 1;
	my $ptr  = $rv;
	my @nodelist;
	
	my $count_lists;
	my $count_tables;
	my $count_figures;

	foreach my $e ($this->{DOM}->getElementsByTagName('*'))
	{
		my $tag = $e->localname;
		
		if ($tag eq 'figure' || $e->getAttribute('class') =~ / (^|\s) figure (\s|$) /x)
		{
			my $node = {};
			bless $node, 'Cognition::DocModel::Figure';
			push @{ $ptr->{kids}->[-1]->{kids} }, $node;
			$node->{tag} = 'figure';
			$node->{count} = ++$count_figures;
			$node->{object} = Cognition::uF::figure::parse($this, $e);
			$node->{_dom} = $e;
			$node->{_page} = $this;
			$node->{up} = $ptr->{kids}->[-1];
			push @nodelist, $node;
		}
		
		elsif ($e->getAttribute('class') =~ / (^|\s) xoxo (\s|$) /x)
		{
			my $node = {};
			bless $node, 'Cognition::DocModel::XOXO';
			push @{ $ptr->{kids}->[-1]->{kids} }, $node;
			$node->{tag} = 'xoxo';
			$node->{count} = ++$count_lists;
			$node->{object} = Cognition::uF::xoxo::parse($this, $e);
			$node->{_dom} = $e;
			$node->{_page} = $this;
			$node->{up} = $ptr->{kids}->[-1];
			$node->{heading} = length $e->getAttribute('title') ? $e->getAttribute('title') : undef;
			push @nodelist, $node;
		}

		elsif ($tag =~ /^h([1-6])$/i)
		{
			my $thisLevel = $1;
			my $node = {};
			bless $node, 'Cognition::DocModel::Section';
			
			while ($thisLevel > $lvl)
			{
				my $parent = $ptr->{kids}->[-1];
				$parent->{up} = $ptr;
				$ptr = $parent;
				$lvl++;
			}
			while ($thisLevel < $lvl)
			{
				$lvl--;
				$ptr = $ptr->{up};
			}
			if ($thisLevel == $lvl)
			{
				push @{ $ptr->{kids} }, $node;
				$node->{up} = $ptr;
			}
			
			$node->{level}   = $thisLevel;
			$node->{tag}     = $tag;
			$node->{_dom}    = $e;
			$node->{_page}   = $this;
			$node->{outline} = Cognition::HTMLParser::Utils::outline_point($tag);
			$node->{heading} = Cognition::HTMLParser::Utils::STRINGIFY($e);
			push @nodelist, $node;
		}
		
		elsif ( $tag eq 'table' && (length $e->getAttribute('summary') || defined $e->getElementsByTagName('caption')->item(0)) )
		{
			my $node = {};
			bless $node, 'Cognition::DocModel::Table';
			if (defined $ptr->{kids}->[-1])
				{ push @{ $ptr->{kids}->[-1]->{kids} }, $node; }
			else
				{ push @{ $ptr->{kids} }, $node; }
			$node->{tag} = $tag;
			$node->{_dom} = $e;
			$node->{summary} = $e->getAttribute('summary');
			$node->{count} = ++$count_tables;
			$node->{_page} = $this;
			$node->{up} = $ptr->{kids}->[-1];
			my @captions = $e->getElementsByTagName('caption');
			if (@captions)
			{
				$node->{heading} = Cognition::HTMLParser::Utils::STRINGIFY($captions[0]);
			}
			push @nodelist, $node;
		}
	}

	my $thisSubject = Cognition::RDFModel::Subject->new($this->abs_url);
	foreach my $N (@nodelist)
	{
		$this->mergeSubject($N->rdf_subject);
		$this->mergeNS($N);
		
		$thisSubject->addVal('http://purl.org/dc/terms/hasPart',
			Cognition::RDFModel::Value->new($N->uri, 1))
			if ($N->{level} == 1);
	}
	$this->mergeSubject($thisSubject);

	$this->{Structure}   = $rv;
	return $rv;	

} #/sub parse_doc_structure


sub parse_external_rdf
{
	my $this = shift;
	
	return unless ($this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/1999/xhtml/vocab#meta'));
		
	foreach my $val ( @{ $this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/1999/xhtml/vocab#meta')->valObjects } )
	{
		my $uri = URI->new($val->toString);
		next if ($this->{options}->{nofollow}->{ $uri->host });
	
		my $request  = HTTP::Request->new(GET => $val->toString);
		my $response = $this->{UA}->request($request);

		if ($response->header('content-type') =~ /(rdf)/i)
		{
			$this->parse_rdf_data(
				$response->content,
				$response->base,
				$response->header('content-type')
				);
		}
	}
} #/sub parse_external_rdf


sub parse_grddl
{
	my $this = shift;

	return unless ($this->{Profiles}->{'http://www.w3.org/2003/g/data-view'}
		|| !$this->{options}->{grddl_strict_profiles} );

	# Get a list of locally defined transformations.
	my @transformations;
	my $prop = $this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/2003/g/data-view#transformation');
	if ($prop)
	{
		foreach my $v (@{$prop->valObjects})
		{
			my $uri = URI->new($v->toString);
			next if ($this->{options}->{nofollow}->{ $uri->host });
			
			push @transformations, $v->toString if ($v->type);
		}
	}
	
	# Get a list of transformations defined in the profiles.
	if ($this->{options}->{grddl_fetch})
	{
		foreach my $p (keys %{$this->{Profiles}})
		{
			my $uri = URI->new($p);
			next if ($this->{options}->{nofollow}->{ $uri->host });
		
			my $request  = HTTP::Request->new(GET => $p);
			my $response = $this->{UA}->request($request);

			if ($response->header('content-type') =~ /(html|xml)/i)
			{
				$this->{Profiles}->{$p} = Cognition::HTMLParser->new(
					$response->content,
					{
						ua        => $this->{UA},
						request   => $request,
						response  => $response,
						p_AUTO    => 0
					});
				my @trans   = Cognition::HTMLParser::Utils::searchRel(
					'profileTransformation', 
					$this->{Profiles}->{$p}->{DOM}
				);
				foreach my $t (@trans)
				{
					my $url = url $t->getAttribute('href'), $response->base;
					push @transformations, $url->abs->as_string
						unless ($this->{options}->{nofollow}->{ $url->host });
				}
			}
		}
	}
	
	my $attr = $this->_fq2pfx('http://www.w3.org/2003/g/data-view#transformation');
	my $attv = $this->{DOM}->documentElement->getAttribute($attr);
	$attv =~ s/(^\s+|\s+$)//g;
	$attv =~ s/\s+/ /g;
	my @attv = split / /, $attv;
	foreach my $a (@attv)
		{ push @transformations, $this->abs_url($a); }
	
	foreach my $t (@transformations)
	{
		my $uri = URI->new($t);
		next if ($this->{options}->{nofollow}->{ $uri->host });
	
		my $request  = HTTP::Request->new(GET => $t);
		my $response = $this->{UA}->request($request);

		if ($response->header('content-type') =~ /(xslt|xml)/i)
		{
			my $parser     = XML::LibXML->new();
			my $xslt       = XML::LibXSLT->new();
			my $style_doc  = $parser->parse_string($response->content);
			my $stylesheet = $xslt->parse_stylesheet($style_doc);
			my $results    = $stylesheet->transform($this->{DOM});
			my $rdfxml     = $stylesheet->output_string($results);
			$this->parse_rdf_data($rdfxml);
		}
	}
}


sub parse_rdf_data
{
	my $this = shift;
	my $rdf  = shift;
	my $base = shift || $this->{RESPONSE}->base;
	my $mime = shift || 'application/rdf+xml';
	
	my $baseobj = new RDF::Redland::URI("$base");
	my $parser  = new RDF::Redland::Parser(undef, $mime);
	my $stream;
	
	if (!eval{$stream = $parser->parse_string_as_stream($rdf, $baseobj);})
	{
		# Funny little way of outputting an error message.
		my $S = Cognition::RDFModel::Subject->new('file:///dev/stderr');
		my $V = Cognition::RDFModel::Value->new("RDF parsing failed for:\n$rdf");
		$S->addVal('http://purl.org/dc/terms/abstract', $V);
		$this->mergeSubject($S);
		return;
	}

	while ($stream && !$stream->end)	
	{
		my $statement = $stream->current;
		
		my $subj = $statement->subject;
		my $prop = $statement->predicate;
		my $obj  = $statement->object;
		
		if ($subj->is_resource && $prop->is_resource)
		{
			my $S = Cognition::RDFModel::Subject->new($subj->uri->as_string);
			my $P = $prop->uri->as_string;
			my $V;
			
			if ($obj->is_resource)
				{ $V = Cognition::RDFModel::Value->new($obj->uri->as_string, 1); }
			else
				{ $V = Cognition::RDFModel::Value->new($obj->literal_value); }

			$S->addVal($P, $V);				
			$this->mergeSubject($S);	
		}
		
		$stream->next;
	}
	
	if ($RDF::Redland::VERSION gt '1.00_06_00')
	{
		my %ns = $parser->namespaces_seen;
		foreach my $k (keys %ns)
		{
			$this->mergeNS($k, $ns{$k}->as_string, "RDF $k", -1, 0);
		}
	}
	else
	{
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\"([^"]*)\"/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'defaultpfx';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix", -1, 0);
		}
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\'([^']*)\'/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'defaultpfx';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix", -1, 0);
		}
	}
} #/sub parse_rdf_data


sub parse_rdf_comments
{
	my $this = shift;
	my $src = $this->{_raw_source} || $this->{Source};

	while ($src =~ /<!--\s*(.*)\s*-->/sg)
	{
		my $comment = $1;
		if ($comment =~ /^[\s\r\n]*<rdf:RDF/)
		{
			$this->parse_rdf_data($comment);
		}
	}
	
} #/sub parse_rdf_comments


sub parse_internal_rdf
{
	my $this = shift;
	my @rdfs = $this->{DOM}->getElementsByTagNameNS("http://www.w3.org/1999/02/22-rdf-syntax-ns#", 'RDF');

	foreach my $rdf (@rdfs)
	{
		$this->parse_rdf_data($rdf->toString);
	}
	
} #/sub parse_internal_rdf


sub parse_profiles
{
	my $this = shift;
	
	foreach my $head ($this->{DOM}->getElementsByTagName('head'))
	{
		my $profile = $head->getAttribute('profile');
		$profile =~ s/(^\s+|\s+$)//g;
		foreach my $url (split /\s+/s, $profile)
		{
			$url = $this->abs_url($url);
			$this->{Profiles}->{ $url } = 1;
		}
	}
	
	foreach my $link (Cognition::HTMLParser::Utils::searchRel('profile', $this->{DOM}))
	{
		my $url = $this->abs_url($link->getAttribute('href'));
		$this->{Profiles}->{ $url } = 1;
	}
	
} #/sub parse_profiles


sub parse_metadata
{
	my $this = shift;
	my $myurl = $this->abs_url;
	
	$this->parse_rfc2731_namespaces;

	# Parse <title> and boring stuff.
	my $S = Cognition::RDFModel::Subject->new($myurl);
	foreach my $t ($this->{DOM}->getElementsByTagName('title'))
	{
		my $val = Cognition::RDFModel::Value->new(Cognition::HTMLParser::Utils::STRINGIFY($t));
		$val->setDom($t);
		$S->addVal('http://purl.org/dc/terms/title', $val);
	}
	my $val = Cognition::RDFModel::Value->new('http://purl.org/dc/dcmitype/Text', 1);
	$val->setVocab('http://purl.org/dc/terms/DCMIType');
	$S->addVal('http://purl.org/dc/terms/type', $val);
	$this->mergeSubject($S);

	$this->parse_profiles;

	$this->parse_metadata_links;
	$this->parse_metadata_metas;
	$this->parse_metadata_roles;
	
	$this->parse_metadata_erdf
		if ($this->{options}->{p_erdf});
	
	if ($this->{options}->{p_rdfa})
	{
		my $doit = 1;
		if ($this->{options}->{rdfa_strict_doctype})
		{
			my $src = $this->{_raw_source} || $this->{Source};
			if ($src !~ m| <!DOCTYPE \s+ html \s+ PUBLIC \s+ "-//W3C//DTD\sXHTML\+RDFa\s1\.0//EN"|six)
			{
				$doit = 0;
			}
		}
		$this->parse_metadata_rdfa if ($doit);
	}
	
	$this->parse_internal_rdf
		if ($this->{options}->{p_rdf});

	$this->parse_external_rdf
		if ($this->{options}->{p_rdfx});

	$this->parse_rdf_comments
		if ($this->{options}->{p_comments});

	# We deliberately parse uF *after* most of the RDF-like stuff. This allows
	# us to use RDFa, etc to select the foaf:primaryTopic.
	$this->parse_uformats
		if ($this->{options}->{p_uf});
	
	$this->parse_doc_structure
		if ($this->{options}->{p_structure});
	
	$this->parse_grddl
		if ($this->{options}->{p_grddl});
		
} #/sub parse_metadata


sub _pfx2fq
{
	my $this = shift;
	my $key  = shift;
	my $mode = shift;
	my $dom  = shift;
	my ($ns, $term, $nsurl);
	
	$mode = '.:' unless length $mode;
	
	if (defined $dom && ($mode eq '.:') && length ($dom->getAttribute('rel').$dom->getAttribute('rev')))
	{
		# From HTML 4.01 and HTML 3.2 specs
		if ($key =~ /^\:?( Alternate | Stylesheet | Start | Next | Prev |
			Contents | Index | Glossary | Chapter | Section | Subsection |
			Appendix | Help | Bookmark | Previous | Search | Made )$/ix)
		{
			$key = "xhv:$key";
		}
		# And the HTML 5 spec.
		elsif ($key =~ /^\:?( alternate | archives | author | bookmark | contact |
			external | feed | first | help | icon | index | last | license | next |
			nofollow | noreferrer | pingback | prefetch | prev | search |
			stylesheet | sidebar | tag | up )$/ix)
		{
			$key = "xhv:$key";
		}
		# Some others. Not sure where they're defined but RDFa test suite mentions.
		elsif ($key =~ /^\:?( cite | meta | p3pv1 )$/ix)
		{
			$key = "xhv:$key";
		}
		elsif ($key eq 'transformation')
		{
			return 'http://www.w3.org/2003/g/data-view#'.$key;
		}
	}
	
	# In "dash mode" allows the hyphen to be used as a namespace
	# prefix separator. This is for eRDF.
	if (($mode eq '-' && $key =~ /^([A-Z_][A-Z0-9_]*)[\.\:-](.+)$/i)
	# Normally, dash is allowed as part of the namespace prefix.    
	|| ($mode eq '.:' && $key =~ /^([A-Z_][A-Z0-9_-]*)[\.\:](.+)$/i))
	{
		$ns = $1;
		$term = $2;
		my $xp = undef;
		$xp = $dom->getAttribute('_xpath') if ($dom);
		
		my $NS = Cognition::Namespace::bestMatch($ns, $xp, @{$this->{Namespaces}});
		$nsurl = $NS->{uri};
		
		# In eRDF, cope with undefined prefixes by assuming that
		# the class name was never intended as eRDF.
		if ($mode eq '-' && !$nsurl)
		{
			return $key;
		}
		
		# Otherwise, if an undefined prefix is used, complain quietly.
		$nsurl = 'http://undefined-namespace-prefix.invalid/' if (!$nsurl);
		
		return $nsurl.$term;
	}

	# If no valid prefix is found, treat as case-insensitive default namespace
	return lc($key);

} #/sub _pfx2fq


sub _fq2pfx
{
	my $this = shift;
	my $key  = shift;
	my $mode = shift || ':';

	$key =~ s/(^\<|\>$)//g;
	$key =~ /^(.*[\#\/])([^\#\/]+)$/;

	my $nsurl = $1;
	$key      = $2;
	my $preferred_pfx;
	
	$key = 'null' unless (length $key);

	if (defined $this->{NamespacePrefixes}->{$nsurl})
		{ return $this->{NamespacePrefixes}->{$nsurl}.$mode.$key; }

	foreach my $ns (@{ $this->{Namespaces} })
	{
		$preferred_pfx = $ns->{prefix}
			if ($ns->{uri} eq $nsurl);
	}

	if ($preferred_pfx)
	{
		my $pfx = $preferred_pfx;
		while (defined $this->{NamespacePrefixesReverse}->{$pfx})
			{ $pfx = $preferred_pfx . int(rand(999)); }

		$this->{NamespacePrefixesReverse}->{$pfx} = $nsurl;
		$this->{NamespacePrefixes}->{$nsurl}      = $pfx;

		return $pfx.$mode.$key;
	}
	
	return $key;

} #/sub _fq2pfx


sub abs_url
{
	my $this  = shift;
	my $param = shift || '';
	my $dom   = shift;
	
	# Support for CURIEs.
	if ($param =~ /^\[([A-Za-z0-9_]+)\:([A-Za-z0-9_]+)\]$/)
	{	
		my $ns    = $1;
		my $curie = $2;

		my $expanded = $this->_pfx2fq("$ns:$curie", undef, $dom);
		return $expanded if (length $expanded);
	}
	
	my $url = url $param, $this->{BASE};
	return $url->abs->as_string;
} #/sub abs_url


sub mergeSubject
{
	my $this    = shift;
	my $mergeme = shift;
	my $target;

	return unless (defined $mergeme);
	
	foreach my $ex (keys %{$this->{RDF}})
	{
		my $existing = $this->{RDF}->{$ex};
		next unless (defined $existing);
		
		if ($existing->equal_to($mergeme))
		{
			$target = $ex;
			last;
		}
	}
	
	if (defined $target)
	{
		$this->{RDF}->{$target}->merge($mergeme);
		$this->{RDF}->{$target}->{page} = $this;
	}
	else
	{
		$this->{RDF}->{ $mergeme->uri } = $mergeme;
		$this->{RDF}->{ $mergeme->uri }->{page} = $this;
	}
}


sub mergeNS
# Can be passed either an object with a metadata_ns method or
# a list of suggestedPrefix, url, title.
{
	my $this    = shift;
	my $mergeme = shift;
	
	if (ref $mergeme && UNIVERSAL::can($mergeme, 'metadata_ns'))
	{
		my $mns = $mergeme->metadata_ns;
		foreach my $k (keys %$mns)
		{
			$this->mergeNS($k, $mns->{$k}->{nsurl}, $mns->{$k}->{title}, undef, 0);
		}
	}
	
	else
	{
		my $suggestedPrefix = $mergeme;
		my $NS = Cognition::Namespace->new($suggestedPrefix, @_);
		
		my $addIt = 1;
		foreach my $existing (@{$this->{Namespaces}})
		{
			if (Cognition::Namespace::equiv($existing, $NS))
			{
				$addIt = 0;
				last;
			}
		}
		
		push @{$this->{Namespaces}}, $NS
			if ($addIt);

		return $suggestedPrefix;
	}
}


sub html2xhtml
{
	my $this = shift;
	my $html = shift;
	
	$html =~ s/\<\!DOCTYPE[^>]*\>//i;
	
	my $t = HTML::TreeBuilder->new();
	$t->implicit_tags(1);
	$t->p_strict(1);
	$t->parse_content($html);
	$t->{'xmlns'} = 'http://www.w3.org/1999/xhtml';
	
	my $xhtml = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
		. "<!DOCTYPE html\n"
		. "\tPUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
		. "\t\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
		. $t->as_XML(undef, "\t", {
			'area'=>1,
			'base'=>1,
			'basefont'=>1,
			'br'=>1,
			'col'=>1,
			'frame'=>1,
			'hr'=>1,
			'img'=>1,
			'input'=>1,
			'isindex'=>1,
			'link'=>1,
			'meta'=>1,
			'param'=>1
		});

	# XHTML now has some annoying named entities.
	#my %ents = %HTML::Entities::entity2char;
	#foreach my $e (keys %ents)
	#{
	#	my $char = $ents{$e};
	#	$e =~ s/\;$//;
	#	$xhtml =~ s/ \& $e \; / sprintf "&#x%X;", ord($char) /gex;
	#}
	
	return encode_entities($xhtml, "\x80-\xFF");
	
	return $xhtml;
} #/sub html2xhtml


1; #/package Cognition::HTMLParser




package Cognition::Namespace;

sub new
{
	my $class = shift;
	my ($prefix, $uri, $title, $scope, $loose) = @_;
	my $this;
	
	$this = {
		prefix  => $prefix,
		uri     => $uri,
		title   => $title,
		scope   => $scope,
		loose   => $loose
	};
	
	bless $this, $class;
}

sub equiv
{
	my $existing = shift;
	my $new      = shift;
	
	if ($existing->{loose})
	{
		return 0 if (lc($existing->{prefix}) ne lc($new->{prefix}));
	}
	else
	{
		return 0 if ($existing->{prefix} ne $new->{prefix});
	}
	
	return 0 if ($existing->{uri} ne $new->{uri});
	return 1 if (!defined $existing->{scope});
	return 1 if ($existing->{scope} == -1);
	
	return 1
		if (((length $existing->{scope}) < (length $new->{scope}))
		&&  substr($new->{scope}, 0, (length $existing->{scope})) eq $existing->{scope});
		
	return 0;
}

sub bestMatch
{
	my $prefix    = shift;
	my $xpath     = shift;	
	my $bestMatch = undef;
	my $highScore = -10;
	
	foreach my $ns (@_)
	{
		my $nsScore = -11;
		if ( ($ns->{scope}<0) || (!defined $ns->{scope}) )
		{
			if ($ns->{loose} && (lc($ns->{prefix}) eq lc($prefix)))
			{
				$nsScore = (defined $ns->{scope}) ? 0 : -2;
			}
			elsif ($ns->{prefix} eq $prefix)
			{
				$nsScore = (defined $ns->{scope}) ? 1 : -1;
			}
		}
		elsif ( (length $xpath)
		&& (substr($xpath,0,(length $ns->{scope})) eq $ns->{scope}) )
		{
			if ($ns->{loose} && (lc($ns->{prefix}) eq lc($prefix)))
			{
				$nsScore = length $ns->{scope};
			}
			elsif ($ns->{prefix} eq $prefix)
			{
				$nsScore = length $ns->{scope};
			}
		}
		
		if ($nsScore > $highScore)
		{
			$highScore = $nsScore;
			$bestMatch = $ns;
		}
	}
	
	return $bestMatch;
}

1;
