#!/usr/bin/perl

######################################################################
package Cognition::uF::hCalendar;
######################################################################

use XML::LibXML qw(:all);
use strict;
use CGI::Util;
use Digest::SHA1 qw(sha1_hex);
use Cognition::uF;
use Cognition::uF::hEvent;
use Cognition::uF::hTodo;
use Cognition::uF::hAlarm;
use Cognition::uF::hFreebusy;

BEGIN
{
	our $VERSION = 0.1;
	our @ISA = qw(Cognition::uF);
}


sub consume
{
	my $page   = shift;

	if ( $page->{Profiles}->{'http://dannyayers.com/microformats/hcalendar-profile'}
	||   $page->{Profiles}->{'http://purl.org/uF/hCalendar/1.0/'}
	||   $page->{Profiles}->{'http://purl.org/uF/2008/03/'}
	||  !$page->{options}->{uf_strict_profiles})
	{	
		$page->{uF}->{hCalendar} = [ Cognition::uF::hCalendar::parse_all($page) ];
		foreach my $hcalendar (@{ $page->{uF}->{hCalendar} })
		{
			$page->mergeSubject( $hcalendar->rdf_subject );
	
			foreach my $childtype (qw(vevent vtodo valarm vfreebusy))
			{
				foreach my $h ( @{ $hcalendar->{$childtype} } )
				{
					$page->mergeSubject( $h->rdf_subject );
					
					foreach my $persontype (qw(contact attendee organizer))
					{
						foreach my $hcard (@{ $h->{$persontype} })
						{
							$page->mergeSubject($hcard->rdf_subject)
								if (UNIVERSAL::can($hcard, 'rdf_subject'));
							
							foreach my $c (@{$hcard->{tel}})
								{ $page->mergeSubject($c->rdf_subject); }
							foreach my $c (@{$hcard->{email}})
								{ $page->mergeSubject($c->rdf_subject); }
							foreach my $c (@{$hcard->{org}})
								{ $page->mergeSubject($c->rdf_subject); }
							foreach my $c (@{$hcard->{impp}})
								{ $page->mergeSubject($c->rdf_subject); }
						}
					}
					
					if ($childtype eq 'vfreebusy')
					{
						foreach my $c (@{$h->{freebusy}})
							{ $page->mergeSubject($c->rdf_subject); }
					}
				}
			}
		}
		if (defined $page->{uF}->{hCalendar}->[0])
			{ $page->mergeNS( $page->{uF}->{hCalendar}->[0] ); }
	}
}


sub parse_all
{
	my $page   = shift;
	my $within = shift || $page->{DOM};
	my $fullPageMode;
	my @nodes  = Cognition::HTMLParser::Utils::searchClass('vcalendar', $within);
	my @rv;
	
	if (!@nodes && $within->nodeType==XML_DOCUMENT_NODE)
	{
		push @nodes, $within->getDocumentElement ;
		$fullPageMode = 1;
	}
	elsif (!@nodes)
	{
		push @nodes, $within;
	}
	
	foreach my $hc (@nodes)
	{
		my $hc_parsed = parse($page, $hc);
		push @rv, $hc_parsed; 
	}
	
	$rv[0]->{_id} = $page->abs_url if ($fullPageMode);
	
	return @rv;
} #/sub parse_all


sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;
	
	if ($this->{_dom}->nodeType == XML_DOCUMENT_NODE || $this->{_dom}->tagName eq 'html')
	{
		push @rv, $this->{_page}->abs_url;
	}
	elsif (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	#push @rv, Cognition::HTMLParser::Utils::XPointer($this->{_dom}, $this->{_page}->abs_url);
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'Calendar')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri


sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2445#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->setDom($this->{_dom});
	
	return $subject
		unless ($this->{vevent}
			|| $this->{vtodo}
			|| $this->{vjournal}
			|| $this->{valarm}
			|| $this->{vfreebusy});
	
	$subject->addObject($this);
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'vcalendar', 1));

	$subject->addVal('http://purl.org/dc/terms/title', 
		Cognition::RDFModel::Value->new($this->{_dom}->getAttribute('title')))
		if (length $this->{_dom}->getAttribute('title'));

	# Patterns for simple data.
	my @_SINGLES  = qw(version prodid x-wr-calname x-wr-caldesc calscale method);
	my @_OBJECTS  = qw(vevent vtodo valarm vfreebusy vjournal);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($ns.$property, $val);
	}

	# Embedded objects which may be represented as resources
	foreach my $property (@_OBJECTS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			my $val;
			if (ref $V)
				{ $val = Cognition::RDFModel::Value->new($V->uri, 1); }
			else
				{ $val = Cognition::RDFModel::Value->new($V); }
			$subject->addVal($ns.$property, $val);
		}
	}
	
	return $subject;
} #/sub rdf_subject


sub metadata_ns
{
	my $this = shift;
	return { 'ical' => { nsurl=>'urn:ietf:rfc:2445#',title=>'iCalendar' } };
} #/sub metadata_ns


sub parse
{
	my $page = shift;
	my $rv = { '_dom'=>shift };
	my $hc_node = $rv->{'_dom'}->cloneNode(1);

	my $id = $hc_node->getAttribute('id');
	$rv->{_id} = $id if (length $id);
	$rv->{_page} = $page;
	
	# Expand microformat patterns.
	my $incl_iterations = 0;
	my $replacements = 1;
	while (($incl_iterations < Cognition::uF::MAX_INCLUDE_ITERATIONS) && $replacements)
	{
		$replacements = Cognition::uF::ufp_include($hc_node, $page) 
		              + Cognition::uF::ufp_include2($hc_node, $page);
		$incl_iterations++;
	}
	Cognition::uF::ufp_abbr($hc_node, $page);
	Cognition::uF::ufp_table($hc_node, $page);
	
	# Pull out nested hEvents
	my @events = Cognition::uF::hEvent::parse_all($page, $hc_node);
	$rv->{vevent} = \@events unless (!@events);

	# Pull out nested hTodo.
	my @events = Cognition::uF::hTodo::parse_all($page, $hc_node);
	$rv->{vtodo} = \@events unless (!@events);
	
	# Pull out nested hAtom Entries.
	my @events = Cognition::uF::hEntry::parse_all($page, $hc_node);
	$rv->{vjournal} = \@events unless (!@events);
	
	# Pull out nested hAlarm.
	my @events = Cognition::uF::hAlarm::parse_all($page, $hc_node);
	$rv->{valarm} = \@events unless (!@events);
	
	# Pull out nested hFreebusy.
	my @events = Cognition::uF::hFreebusy::parse_all($page, $hc_node);
	$rv->{vfreebusy} = \@events unless (!@events);

	my %xpath;
	foreach my $component (qw(vevent vtodo vjournal valarm vfreebusy))
	{
		foreach my $object (@{ $rv->{$component} })
		{
			my $xp = $object->{'_dom'}->getAttribute('_xpath');
			$xpath{$xp} = $object;
		}
	}
	my @xpaths = keys %xpath;
	foreach my $xp (@xpaths)
	{
		unless (defined $xpath{$xp}->{'parent'})
		{
			my $parent = _findParent($xp, @xpaths);
			if ($parent)
			{
				$xpath{$xp}->{'parent'} = $xpath{$parent}->uri;
				push @{ $xpath{$parent}->{'child'} }, $xpath{$xp}->uri;
			}
		}
	}

	
	# Now that we have reached here, no other composite microformats have any
	# business being nested within this root element. So in the interest of
	# Microformat opacity, let's destroy them.
	Cognition::uF::destroyer($hc_node);
	
	# Simple attributes - singles
	my @simples = qw(version prodid x-wr-calname x-wr-caldesc calscale method);	
	foreach my $simple (@simples)
	{
		my @value_nodes = Cognition::HTMLParser::Utils::searchClass($simple, $hc_node);
		my $value_node  = $value_nodes[0];
		next unless ($value_node);
		my $value = Cognition::HTMLParser::Utils::STRINGIFY($value_node, 'value');
		$rv->{$simple} = $value;
	}
	
	bless $rv;
	return $rv;
	
} #/sub parse


sub _findParent
{
	my $x = shift;
	my $longest = '';
	
	foreach my $potential (@_)
	{
		if (_ancestorOf($potential, $x))
		{
			$longest = $potential
				if (length($potential) > length($longest));
		}
	}
	
	return $longest;
}

sub _ancestorOf
{
	my ($a, $b) = @_;
	return if ($a eq $b);
	return (substr($b, 0, length($a)) eq $a);
}


1; #/package Cognition::uF::hCalendar




